/*
    CHAOS target for iptables

    Copyright © Jan Engelhardt <jengelh [at] gmx de>, 2006 - 2007
    released under the terms of the GNU General Public
    License version 2.x and only versions 2.x.
*/
#include <getopt.h>
#include <stdio.h>
#include <string.h>

#include <iptables.h>
#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter/xt_CHAOS.h>

static void libipt_chaos_help(void)
{
	printf(
		"CHAOS target v%s options:\n"
		"  --delude     Enable DELUDE processing for TCP\n"
		"  --tarpit     Enable TARPIT processing for TCP\n",
		IPTABLES_VERSION);
	return;
}

static int libipt_chaos_parse(int c, char **argv, int invert,
    unsigned int *flags, const struct ipt_entry *entry,
    struct ipt_entry_target **target)
{
	struct xt_chaos_info *info = (void *)((*target)->data);
	switch(c) {
		case 'd':
			info->variant = XTCHAOS_DELUDE;
			*flags |= 0x02;
			return 1;
		case 't':
			info->variant = XTCHAOS_TARPIT;
			*flags |= 0x01;
			return 1;
	}
	return 0;
}

static void libipt_chaos_check(unsigned int flags)
{
	if(flags != 0x03)
		return;
	/* If flags == 0x03, both were specified, which should not be. */
	exit_error(PARAMETER_PROBLEM,
	           "CHAOS: only one of --tarpit or --delude may be specified");
	return;
}

static void libipt_chaos_print(const struct ipt_ip *ip,
    const struct ipt_entry_target *target, int numeric)
{
	const struct xt_chaos_info *info = (const void *)target->data;
	switch(info->variant) {
		case XTCHAOS_DELUDE:
			printf("DELUDE ");
			break;
		case XTCHAOS_TARPIT:
			printf("TARPIT ");
			break;
		default:
			break;
	}
	return;
}

static void libipt_chaos_save(const struct ipt_ip *ip,
    const struct ipt_entry_target *target)
{
	const struct xt_chaos_info *info = (const void *)target->data;
	switch(info->variant) {
		case XTCHAOS_DELUDE:
			printf("--delude ");
			break;
		case XTCHAOS_TARPIT:
			printf("--tarpit ");
			break;
		default:
			break;
	}
	return;
}

static struct option libipt_chaos_opts[] = {
	{"delude", 0, NULL, 'd'},
	{"tarpit", 0, NULL, 't'},
	{NULL},
};

static struct iptables_target libipt_chaos_info = {
	.name          = "CHAOS",
	.version       = IPTABLES_VERSION,
	.size          = IPT_ALIGN(sizeof(struct xt_chaos_info)),
	.userspacesize = IPT_ALIGN(sizeof(struct xt_chaos_info)),
	.help          = libipt_chaos_help,
	.parse         = libipt_chaos_parse,
	.final_check   = libipt_chaos_check,
	.print         = libipt_chaos_print,
	.save          = libipt_chaos_save,
	.extra_opts    = libipt_chaos_opts,
};

static __attribute__((constructor)) void libipt_chaos_init(void)
{
	register_target(&libipt_chaos_info);
	return;
}
